/*
 Copyright (c) 2009-2010 Alessandro Sivieri <alessandro.sivieri@gmail.com>

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Library General Public
 License version 2 as published by the Free Software Foundation.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Library General Public License for more details.

 You should have received a copy of the GNU Library General Public License
 along with this library; see the file COPYING.LIB.  If not, write to
 the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 Boston, MA 02110-1301, USA.
*/

#include <QMetaObject>
#include <QStringList>

#include <KDebug>

#include <Soprano/Model>
#include <Soprano/Node>
#include <Soprano/QueryResultIterator>
#include <Soprano/Vocabulary/NAO>
#include <Soprano/Vocabulary/RDFS>
#include <Soprano/Vocabulary/XMLSchema>
#include <Nepomuk/ResourceManager>
#include <Nepomuk/Resource>
#include <Nepomuk/Types/Class>
#include <Nepomuk/Query/QueryParser>
#include <Nepomuk/Query/Query>
#include <Nepomuk/Query/Term>
#include <Nepomuk/Query/ResourceTypeTerm>
#include <Nepomuk/Query/AndTerm>
#include <Nepomuk/Query/ComparisonTerm>
#include <Nepomuk/Query/LiteralTerm>
#include <Nepomuk/Query/ResourceTerm>

#include "tagsfacet.h"
#include "nie.h"
#include "facetmodel.h"
#include "facetfilterproxysortmodel.h"
#include "facetspanel.h"

using namespace Nepomuk;

class TagsFacet::Private
{
    public:
        Private(FacetsPanel* parent): panel(parent)
        {
        }
        QString folder;
        FacetsPanel* panel;
        QStringList labels;
};

TagsFacet::TagsFacet(QWidget* parent): Facet(parent), d(new Private(qobject_cast<FacetsPanel*>(parent)))
{
    loadTags();
}

TagsFacet::~TagsFacet()
{
    delete d;
}

void TagsFacet::loadTags()
{
    emit clear();

    // ok, now we are getting the true resources query
    /*Query::Query query = d->panel->getQuery();
    query.addRequestProperty(Query::Query::RequestProperty(Types::Property(Soprano::Vocabulary::NAO::hasTag()), false));
    Soprano::Util::AsyncQuery* asyncQuery = Soprano::Util::AsyncQuery::executeQuery(Nepomuk::ResourceManager::instance()->mainModel(),
                                                                               query.toSparqlQuery(),
                                                                               Soprano::Query::QueryLanguageSparql);*/
    // get all possible tags
    QString query = QString("select distinct ?tag ?label where { ?r %1 ?tag . ?tag %2 ?label . }")
            .arg(Soprano::Node::resourceToN3(Soprano::Vocabulary::NAO::hasTag()),
                 Soprano::Node::resourceToN3(Soprano::Vocabulary::NAO::identifier()));
    Soprano::Util::AsyncQuery* asyncQuery = Soprano::Util::AsyncQuery::executeQuery(Nepomuk::ResourceManager::instance()->mainModel(),
                                                                               query,
                                                                               Soprano::Query::QueryLanguageSparql);
    connect(asyncQuery, SIGNAL(nextReady(Soprano::Util::AsyncQuery*)), this, SLOT(queryNextResult(Soprano::Util::AsyncQuery*)));
    connect(asyncQuery, SIGNAL(finished(Soprano::Util::AsyncQuery*)), this, SLOT(queryFinished(Soprano::Util::AsyncQuery*)));

    kDebug() << query;
}

void TagsFacet::loadFrequencies()
{
    QString query = QString("select ?label count(?t) as ?freq where { graph ?g { ?r %1 ?t . } . ?g %2 ?date . ?t %3 ?label . } ORDER BY DESC(?freq)")
            .arg(Soprano::Node::resourceToN3(Soprano::Vocabulary::NAO::hasTag()),
                 Soprano::Node::resourceToN3(Soprano::Vocabulary::NAO::created()),
                 Soprano::Node::resourceToN3(Soprano::Vocabulary::NAO::identifier()));
    Soprano::Util::AsyncQuery* asyncQuery = Soprano::Util::AsyncQuery::executeQuery(Nepomuk::ResourceManager::instance()->mainModel(),
                                                                               query,
                                                                               Soprano::Query::QueryLanguageSparql);
    connect(asyncQuery, SIGNAL(nextReady(Soprano::Util::AsyncQuery*)), this, SLOT(queryFreqNextResult(Soprano::Util::AsyncQuery*)));
    connect(asyncQuery, SIGNAL(finished(Soprano::Util::AsyncQuery*)), this, SLOT(queryFreqFinished(Soprano::Util::AsyncQuery*)));

    kDebug() << query;
}

void TagsFacet::queryNextResult(Soprano::Util::AsyncQuery* query)
{
    QString label;
    QUrl url;
    int count = 0;

    // get uri and label
    Soprano::Node node1 = query->binding("tag");
    Soprano::Node node2 = query->binding("label");
    if (node1.isValid() && node2.isValid()) {
        url = node1.uri();
        label = node2.literal().toString();
    }
    else {
        kDebug() << "Something wrong: nodes are not valid.";
        query->next();
    }

    // get count
    Query::Term term = Query::ComparisonTerm(Soprano::Vocabulary::NAO::hasTag(), Query::ResourceTerm(url));
    Query::Term mainTerm = d->panel->getTerm();
    if ( mainTerm.isValid() )
        term = Query::AndTerm( mainTerm, term );
    Query::Query q = d->panel->getQuery(term);
    Soprano::QueryResultIterator it = Nepomuk::ResourceManager::instance()->mainModel()->executeQuery(q.toSparqlQuery(Query::Query::CreateCountQuery),
                                                                                                      Soprano::Query::QueryLanguageSparql);
    if ( it.next() ) {
        Soprano::Node node2 = it.binding(0);
        if (node2.isValid()) {
            count = node2.literal().toInt();
        }
    }

    // emit the new tag and invoke the next
    emit newTag(label, count);
    query->next();
}

void TagsFacet::queryFinished(Soprano::Util::AsyncQuery* query)
{
    Q_UNUSED(query);

    kDebug() << "Tags loading finished";
    loadFrequencies();
}

void TagsFacet::queryFreqNextResult(Soprano::Util::AsyncQuery* query)
{
    QString label;
    int frequency = 0;

    // get label
    Soprano::Node node1 = query->binding("freq");
    Soprano::Node node2 = query->binding("label");
    if (node1.isValid() && node2.isValid()) {
        frequency = node1.literal().toInt();
        label = node2.literal().toString();
        emit updateFrequency(label, frequency);
    }
    else {
        kDebug() << "Something wrong: nodes are not valid.";
    }

    query->next();
}

void TagsFacet::queryFreqFinished(Soprano::Util::AsyncQuery* query)
{
    Q_UNUSED(query);

    kDebug() << "Tags frequency obtained";
}

void TagsFacet::doFolderChanged(const QString& folder)
{
    // we want to save the last visited "real" folder
    if (!folder.startsWith("nepomuksearch")) {
        d->folder = folder;
    }
    loadTags();
}

void TagsFacet::doReload()
{
    loadTags();
}

void TagsFacet::emitTerm(QStringList labels)
{
    Query::AndTerm term;
    foreach (QString label, labels) {
        Query::ComparisonTerm ct(Soprano::Vocabulary::NAO::hasTag(), Query::LiteralTerm(label));
        term.addSubTerm(ct);
    }
    emitQueryTermChanged(term);
}

QString TagsFacet::merge(const QStringList& list) const
{
    int i;
    QString result;

    for (i = 0; i < list.size(); ++i) {
        if (i > 0) {
            result.append(" AND ");
        }
        result.append("hastag:");
        result.append(list[i]);
    }

    return result;
}

// why the hell the name is changed?
#include "moc_tagsfacet.cpp"
